/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.email;

import static de.ozgcloud.nachrichten.email.MailSendRequestTestFactory.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import jakarta.activation.DataSource;
import jakarta.mail.MessagingException;
import jakarta.mail.internet.MimeMessage;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.mail.MailSendException;
import org.springframework.mail.javamail.JavaMailSender;
import org.springframework.mail.javamail.MimeMessageHelper;

import lombok.SneakyThrows;

class EMailServiceTest {

	@Spy // NOSONAR
	@InjectMocks
	private EMailService service;

	@Mock
	private JavaMailSender sender;
	@Mock
	private ApplicationEventPublisher publisher;

	@Nested
	class TestSendingEmail {

		@Nested
		class TestSendEmail {
			private MailSendRequest request = MailSendRequestTestFactory.create();

			@Mock // NOSONAR
			private MimeMessage message;

			@BeforeEach
			void initTest() {
				when(sender.createMimeMessage()).thenReturn(message);
			}

			@Test
			void shouldWriteEmail() {
				service.sendEmail(request);

				verify(service).writeEmail(any(), eq(request));
			}

			@Test
			void shouldAddAttachments() {
				service.sendEmail(request);

				verify(service).addAttachments(any(), eq(request));
			}

			@Test
			void shouldSendMessage() {
				service.sendEmail(request);

				verify(service).send(message, request);
			}
		}

		@Nested
		class TestWriteEmail {

			@Mock // NOSONAR
			private MimeMessageHelper helper;

			private MailSendRequest request = MailSendRequestTestFactory.create();

			@Test
			void shouldSetToAddress() throws MessagingException {
				service.writeEmail(helper, request);

				verify(helper).setTo(new String[] { formatMailRecipient(EMAIL_RECIPIENT) });
			}

			@SneakyThrows
			@Test
			void shouldFormatToEmailAddresses() {
				var request = MailSendRequestTestFactory.createBuilder().clearToAddresses().toAddress(EMAIL_RECIPIENT).build();

				service.writeEmail(helper, request);

				verify(helper).setTo(new String[] { formatMailRecipient(EMAIL_RECIPIENT) });
			}

			@SneakyThrows
			@Test
			void shouldFormatFromEmailAddresses() {
				service.writeEmail(helper, request);

				verify(helper).setFrom(FROM);
			}

			@Test
			void shouldSetFromAddress() throws MessagingException {
				service.writeEmail(helper, request);

				verify(helper).setFrom(MailSendRequestTestFactory.FROM);
			}

			@Test
			void shouldSetSubject() throws MessagingException {
				service.writeEmail(helper, request);

				verify(helper).setSubject(MailSendRequestTestFactory.SUBJECT);
			}

			@Test
			void shouldSetBody() throws MessagingException {
				service.writeEmail(helper, request);

				verify(helper).setText(MailSendRequestTestFactory.BODY);
			}

			@Test
			void shouldSetReplyTo() throws MessagingException {
				service.writeEmail(helper, request);

				verify(helper).setReplyTo(MailSendRequestTestFactory.REPLY_TO);
			}

			public static String formatMailRecipient(MailRecipient mailRecipient) {
				return EMAIL_ADDRESS_TEMPLATE.formatted(mailRecipient.firstName(), mailRecipient.lastName(), mailRecipient.email());
			}
		}

		@Nested
		class TestAddAttachments {

			@Mock // NOSONAR
			private MimeMessageHelper helper;

			private MailSendRequest request = MailSendRequestTestFactory.create();

			@Test
			void shouldAddInlineAttachment() throws MessagingException {
				service.addAttachments(helper, request);

				verify(helper).addAttachment(eq(MailSendRequestTestFactory.ATTACHMENT_NAME), any(DataSource.class));
			}
		}

		@Nested
		class TestAddHeaders {
			@Mock // NOSONAR
			private MimeMessage message;

			private MailSendRequest request = MailSendRequestTestFactory.createBuilder().receiptRequired(true).build();

			@Test
			void shouldRequestReceipt() {
				service.prepareAndSendMail(message, request);

				verify(service).requestReceipt(any(), any());
			}

			@Test
			void shouldNotRequestReceipt() {
				MailSendRequest request = MailSendRequestTestFactory.createBuilder().receiptRequired(false).build();

				service.prepareAndSendMail(message, request);

				verify(service, never()).requestReceipt(any(), any());
			}

			@Nested
			class RequestReceipt {

				@Test
				void shouldAddReadReceiptToHeader() throws MessagingException {
					service.requestReceipt(message, request);

					verify(message).addHeader(EMailService.HEADER_READ_RECEIPT, MailSendRequestTestFactory.FROM);
				}

				@Test
				void shouldAddConfirmReadingHeader() throws MessagingException {
					service.requestReceipt(message, request);

					verify(message).addHeader(EMailService.HEADER_CONFIRM_READING, MailSendRequestTestFactory.FROM);
				}

				@Test
				void shouldAddDispositionNotificationHeader() throws MessagingException {
					service.requestReceipt(message, request);

					verify(message).addHeader(EMailService.HEADER_DISPOSITION_NOTIFICATION, MailSendRequestTestFactory.FROM);
				}
			}
		}

		@Nested
		class TestSend {

			@Mock // NOSONAR
			private MimeMessage message;
			private MailSendRequest request = MailSendRequestTestFactory.create();

			@Test
			void shouldCallSender() {
				service.send(message, request);

				verify(sender).send(message);
			}

			@Test
			void shouldTriggerEvent() {
				service.send(message, request);

				verify(publisher).publishEvent(any(MailSentEvent.class));
			}

			@Test
			void shouldTriggerEventOnError() {
				doThrow(MailSendException.class).when(sender).send(any(MimeMessage.class));

				service.send(message, request);

				verify(publisher).publishEvent(any(MailSendErrorEvent.class));
				verify(publisher, never()).publishEvent(any(MailSentEvent.class));
			}
		}
	}
}
